try:
    from pylibdmtx import pylibdmtx
except ImportError:
    pylibdmtx = None
    __all__ = ()
else:
    __all__=('DataMatrix',)

from reportlab.graphics.barcode.common import Barcode
from reportlab.lib.utils import asBytes
from reportlab.platypus.paraparser import _num as paraparser_num
from reportlab.graphics.widgetbase import Widget
from reportlab.lib.validators import isColor, isString, isColorOrNone, isNumber, isBoxAnchor
from reportlab.lib.attrmap import AttrMap, AttrMapValue
from reportlab.lib.colors import toColor
from reportlab.graphics.shapes import Group, Rect

def _numConv(x):
    return x if isinstance(x,(int,float)) else paraparser_num(x)

class _DMTXCheck(object):
    @classmethod
    def pylibdmtx_check(cls):
        if not pylibdmtx:
            raise ValueError('The %s class requires package pylibdmtx' % cls.__name__)

class DataMatrix(Barcode,_DMTXCheck):
    def __init__(self, value='', **kwds):
        self.pylibdmtx_check()
        self._recalc = True
        self.value = value
        self.cellSize = kwds.pop('cellSize','5x5')
        self.size = kwds.pop('size','SquareAuto')
        self.encoding = kwds.pop('encoding','Ascii')
        self.anchor = kwds.pop('anchor','sw')
        self.color = kwds.pop('color',(0,0,0))
        self.bgColor = kwds.pop('bgColor',None)
        self.x = kwds.pop('x',0)
        self.y = kwds.pop('y',0)
        self.border = kwds.pop('border',5)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self,v):
        self._value = asBytes(v)
        self._recalc = True

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self,v):
        self._size = self._checkVal('size', v, pylibdmtx.ENCODING_SIZE_NAMES)
        self._recalc = True

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self,v):
        self._border = _numConv(v)
        self._recalc = True

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self,v):
        self._x = _numConv(v)
        self._recalc = True

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self,v):
        self._y = _numConv(v)
        self._recalc = True

    @property
    def cellSize(self):
        return self._cellSize

    @size.setter
    def cellSize(self,v):
        self._cellSize = v
        self._recalc = True

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self,v):
        self._encoding = self._checkVal('encoding', v, pylibdmtx.ENCODING_SCHEME_NAMES)
        self._recalc = True

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self,v):
        self._anchor = self._checkVal('anchor', v, ('n','ne','e','se','s','sw','w','nw','c'))
        self._recalc = True

    def recalc(self):
        if not self._recalc: return
        data = self._value
        size = self._size
        encoding = self._encoding
        e = pylibdmtx.encode(data, size=size, scheme=encoding)
        iW = e.width
        iH = e.height
        p = e.pixels
        iCellSize = 5
        bpp = 3 #bytes per pixel
        rowLen = iW*bpp
        cellLen = iCellSize*bpp
        assert len(p)//rowLen == iH
        matrix = list(filter(None,
                            (''.join(
                                (('x' if p[j:j+bpp] != b'\xff\xff\xff' else ' ')
                                for j in range(i,i+rowLen,cellLen))).strip()
                            for i in range(0,iH*rowLen,rowLen*iCellSize))))
        self._nRows = len(matrix)
        self._nCols = len(matrix[-1])
        self._matrix = '\n'.join(matrix)

        cellWidth = self._cellSize
        if cellWidth:
            cellWidth = cellWidth.split('x')
            if len(cellWidth)>2:
                raise ValueError('cellSize needs to be distance x distance not %r' % self._cellSize)
            elif len(cellWidth)==2:
                cellWidth, cellHeight = cellWidth
            else:
                cellWidth = cellHeight = cellWidth[0]
            cellWidth = _numConv(cellWidth)
            cellHeight = _numConv(cellHeight)
        else:
            cellWidth = cellHeight = iCellSize
        self._cellWidth = cellWidth
        self._cellHeight = cellHeight
        self._recalc = False
        self._bord = max(self.border,cellWidth,cellHeight)
        self._width = cellWidth*self._nCols + 2*self._bord
        self._height = cellHeight*self._nRows + 2*self._bord

    @property
    def matrix(self):
        self.recalc()
        return self._matrix

    @property
    def width(self):
        self.recalc()
        return self._width

    @property
    def height(self):
        self.recalc()
        return self._height

    @property
    def cellWidth(self):
        self.recalc()
        return self._cellWidth

    @property
    def cellHeight(self):
        self.recalc()
        return self._cellHeight

    def draw(self):
        self.recalc()
        canv = self.canv
        w = self.width
        h = self.height
        x = self.x
        y = self.y
        b = self._bord

        anchor = self.anchor
        if anchor in ('nw','n','ne'):
            y -= h
        elif anchor in ('c','e','w'):
            y -= h//2
        if anchor in ('ne','e','se'):
            x -= w
        elif anchor in ('n','c','s'):
            x -= w//2

        canv.saveState()
        if self.bgColor:
            canv.setFillColor(toColor(self.bgColor))
            canv.rect(x, y-h, w, h, fill=1, stroke=0)
        canv.setFillColor(toColor(self.color))
        canv.setStrokeColor(None)

        cellWidth = self.cellWidth
        cellHeight = self.cellHeight
        yr = y - b - cellHeight
        x += b
        for row in self.matrix.split('\n'):
            xr = x 
            for c in row:
                if c=='x':
                    canv.rect(xr, yr, cellWidth, cellHeight, fill=1, stroke=0)
                xr += cellWidth
            yr -= cellHeight
        canv.restoreState()
    

class DataMatrixWidget(Widget,_DMTXCheck):
    codeName = "DataMatrix"
    _attrMap = AttrMap(
        BASE = Widget,
        value = AttrMapValue(isString, desc='Datamatrix data'),
        x = AttrMapValue(isNumber, desc='x-coord'),
        y = AttrMapValue(isNumber, desc='y-coord'),
        color = AttrMapValue(isColor, desc='foreground color'),
        bgColor = AttrMapValue(isColorOrNone, desc='background color'),
        encoding = AttrMapValue(isString, desc='encoding'),
        size = AttrMapValue(isString, desc='size'),
        cellSize = AttrMapValue(isString, desc='cellSize'),
        anchor = AttrMapValue(isBoxAnchor, desc='anchor pooint for x,y'),
        )

    _defaults = dict(
        x = ('0',_numConv),
        y = ('0',_numConv),
        color = ('black',toColor),
        bgColor = (None,lambda _: toColor(_) if _ is not None else _),
        encoding = ('Ascii',None),
        size = ('SquareAuto',None),
        cellSize = ('5x5',None),
        anchor = ('sw', None),
        )
    def __init__(self,value='Hello Cruel World!', **kwds):
        self.pylibdmtx_check()
        self.value = value
        for k,(d,c) in self._defaults.items():
            v = kwds.pop(k,d)
            if c: v = c(v)
            setattr(self,k,v)

    def rect(self, x, y, w, h, fill=1, stroke=0):
        self._gadd(Rect(x,y,w,h,strokeColor=None,fillColor=self._fillColor))

    def saveState(self,*args,**kwds):
        pass

    restoreState = setStrokeColor = saveState

    def setFillColor(self,c):
        self._fillColor = c

    def draw(self):
        m = DataMatrix(value=self.value,**{k: getattr(self,k) for k in self._defaults})
        m.canv = self
        m.y += m.height
        g = Group()
        self._gadd = g.add
        m.draw()
        return g



#
# Copyright (c) 1996-2000 Tyler C. Sarna <tsarna@sarna.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#      This product includes software developed by Tyler C. Sarna.
# 4. Neither the name of the author nor the names of contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
__all__ = tuple('''registerWidget getCodes getCodeNames createBarcodeDrawing createBarcodeImageInMemory'''.split())
__version__ = '0.9'
__doc__='''Popular barcodes available as reusable widgets'''

_widgets = []
def registerWidget(widget):
    _widgets.append(widget)

def _reset():
    _widgets[:] = []
    from reportlab.graphics.barcode.widgets import BarcodeI2of5, BarcodeCode128, BarcodeStandard93,\
                        BarcodeExtended93, BarcodeStandard39, BarcodeExtended39,\
                        BarcodeMSI, BarcodeCodabar, BarcodeCode11, BarcodeFIM,\
                        BarcodePOSTNET, BarcodeUSPS_4State, BarcodeCode128Auto, BarcodeECC200DataMatrix

    #newer codes will typically get their own module
    from reportlab.graphics.barcode.eanbc import Ean13BarcodeWidget, Ean8BarcodeWidget, UPCA, Ean5BarcodeWidget, ISBNBarcodeWidget
    from reportlab.graphics.barcode.qr import QrCodeWidget
    for widget in (BarcodeI2of5,
                BarcodeCode128,
                BarcodeCode128Auto,
                BarcodeStandard93,
                BarcodeExtended93,
                BarcodeStandard39,
                BarcodeExtended39,
                BarcodeMSI,
                BarcodeCodabar,
                BarcodeCode11,
                BarcodeFIM,
                BarcodePOSTNET,
                BarcodeUSPS_4State,
                Ean13BarcodeWidget,
                Ean8BarcodeWidget,
                UPCA,
                Ean5BarcodeWidget,
                ISBNBarcodeWidget,
                QrCodeWidget,
                BarcodeECC200DataMatrix,
                ):
        registerWidget(widget)
        from reportlab.graphics.barcode import dmtx
        if dmtx.pylibdmtx:
            registerWidget(dmtx.DataMatrixWidget)

_reset()
from reportlab.rl_config import register_reset
register_reset(_reset)

def getCodes():
    """Returns a dict mapping code names to widgets"""
    #the module exports a dictionary of names to widgets, to make it easy for
    #apps and doc tools to display information about them.
    codes = {}
    for widget in _widgets:
        codeName = widget.codeName
        codes[codeName] = widget

    return codes

def getCodeNames():
    """Returns sorted list of supported bar code names"""
    return sorted(getCodes().keys())

def createBarcodeDrawing(codeName, **options):
    """This creates and returns a drawing with a barcode.
    """    
    from reportlab.graphics.shapes import Drawing, Group

    codes = getCodes()
    bcc = codes[codeName]
    width = options.pop('width',None)
    height = options.pop('height',None)
    isoScale = options.pop('isoScale',0)
    kw = {}
    for k,v in options.items():
        if k.startswith('_') or k in bcc._attrMap: kw[k] = v
    bc = bcc(**kw)


    #Robin's new ones validate when setting the value property.
    #Ty Sarna's old ones do not.  We need to test.
    if hasattr(bc, 'validate'):
        bc.validate()   #raise exception if bad value
        if not bc.valid:
            raise ValueError("Illegal barcode with value '%s' in code '%s'" % (options.get('value',None), codeName))

    #size it after setting the data    
    x1, y1, x2, y2 = bc.getBounds()
    w = float(x2 - x1)
    h = float(y2 - y1)
    sx = width not in ('auto',None)
    sy = height not in ('auto',None)
    if sx or sy:
        sx = sx and width/w or 1.0
        sy = sy and height/h or 1.0
        if isoScale:
            if sx<1.0 and sy<1.0:
                sx = sy = max(sx,sy)
            else:
                sx = sy = min(sx,sy)

        w *= sx
        h *= sy
    else:
        sx = sy = 1

    #bc.x = -sx*x1
    #bc.y = -sy*y1
    d = Drawing(width=w,height=h,transform=[sx,0,0,sy,-sx*x1,-sy*y1])
    d.add(bc, "_bc")
    return d

def createBarcodeImageInMemory(codeName,**options):
    """This creates and returns barcode as an image in memory.
    Takes same arguments as createBarcodeDrawing and also an
    optional format keyword which can be anything acceptable
    to Drawing.asString eg gif, pdf, tiff, py ......
    """
    format = options.pop('format','png')
    d = createBarcodeDrawing(codeName, **options)
    return d.asString(format)



__all__ = '''MultiCol'''.split()
from reportlab.lib.utils import strTypes
from .flowables import Flowable, _Container, _FindSplitterMixin, _listWrapOn

class MultiCol(_Container,_FindSplitterMixin,Flowable):
	def __init__(self,contents,widths, minHeightNeeded=36, spaceBefore=None, spaceAfter=None):
		if len(contents)!=len(widths):
			raise ValueError('%r len(contents)=%d not the same as len(widths)=%d' % (self,len(contents),len(widths)))
		self.contents = contents
		self.widths = widths
		self.minHeightNeeded = minHeightNeeded
		self._spaceBefore = spaceBefore
		self._spaceAfter = spaceAfter
		self._naW = None

	def nWidths(self,aW):
		if aW==self._naW: return self._nW
		nW = [].append
		widths = self.widths
		s = 0.0
		for i,w in enumerate(widths):
			if isinstance(w,strTypes):
				w=w.strip()
				pc = w.endswith('%')
				if pc: w=w[:-1]
				try:
					w = float(w)
				except:
					raise ValueError('%s: nWidths failed with value %r' % (self,widths[i]))
				if pc: w = w*0.01*aW
			elif not isinstance(w,(float,int)):
				raise ValueError('%s: nWidths failed with value %r' % (self,widths[i]))

			s += w
			nW(w)

		self._naW = aW
		s = aW / s
		self._nW = [w*s for w in nW.__self__]
		return self._nW

	def wrap(self,aW,aH):
		widths = self.nWidths(aW)
		w = h = 0.0
		canv = self.canv
		h = 0
		for faW,F in zip(widths,self.contents):
			if not F:
				fW = faW
				fH = 0
			else:
				fW,fH = _listWrapOn(F,faW,canv)
			h = max(h,fH)
			w += fW
		self.width = w
		self.height = h
		return w, h

	def split(self,aW,aH):
		if aH<self.minHeightNeeded:
			return []
		widths = self.nWidths(aW)
		S = [[],[]]
		canv = self.canv
		for faW,F in zip(widths,self.contents):
			if not F:
				fW = faW
				fH0 = 0
				S0 = []
				S1 = []
			else:
				fW,fH0,S0,S1 = self._findSplit(canv,faW,aH,content=F,paraFix=False)
				if S0 is F: return [] #we failed to find a split
			S[0].append(S0)
			S[1].append(S1)

		return	[
				MultiCol(S[0],
					self.widths,
					minHeightNeeded=self.minHeightNeeded,
					spaceBefore=self._spaceBefore,
					spaceAfter=self._spaceAfter),
				MultiCol(S[1],
					self.widths,
					minHeightNeeded=self.minHeightNeeded,
					spaceBefore=self._spaceBefore,
					spaceAfter=self._spaceAfter),
				]

	def getSpaceAfter(self):
		m = self._spaceAfter
		if m is None:
			m = 0
			for F in self.contents:
				m = max(m,_Container.getSpaceAfter(self,F))
		return m

	def getSpaceBefore(self):
		m = self._spaceBefore
		if m is None:
			m = 0
			for F in self.contents:
				m = max(m,_Container.getSpaceBefore(self,F))
		return m

	def drawOn(self, canv, x, y, _sW=0):
		widths = self._nW
		xOffs = 0
		for faW,F in zip(widths,self.contents):
			_Container.drawOn(self, canv, x+xOffs, y, content=F, aW=faW)
			xOffs += faW



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/platypus/__init__.py
__version__='3.5.20'
__doc__='''Page Layout and Typography Using Scripts" - higher-level framework for flowing documents'''

from .flowables import *
from .frames import *
from .multicol import *
from .paragraph import *
from .paraparser import *
from .tables import *
from .doctemplate import *
from .xpreformatted import *
